         .586p
DGROUP group _DATA,_BSS
_TEXT  segment dword public use32 'CODE'
       assume cs:_TEXT, ds:DGROUP, ss:DGROUP

;-------------------------------
; Fellow Amiga emulator
; Petter Schau 1996/97
; Assembler part of Soundblaster
; specific sound driver
;-------------------------------


;----------------------------------------------------------
; Routines to put samples generated by the emulation into a
; dmabuffer
;----------------------------------------------------------

; A pointer to one of these routines are stored in
; _audio_putinbuffer, and it is called when the emulation has produced
; some new samples, the number of samples produced are in ebx.
; The samples produced are found in two arrays called _left and _right
; These samples are always 16-bit signed samples.

; These routines takes those samples and stores them in a dma-buffer
; in the format soundblaster uses.

; There are 4 cases which must be handled for SB:
; 1. In 16-bit stereo samples are simply moved and interleaved to the buffer.
; 2. In 16-bit mono samples from _left and _right are added and stored in the
;    buffer
; 3. In 8-bit stereo samples are converted to unsigned samples, shifted
;    down 8 bits and stored and interleaved in the dma-buffer.
; 4. In 8-bit mono samples from _left and _right are added, converted to
;    unsigned and then moved to the buffer.

; Also, zero samples can be generated :-)

; ebx - number of samples generated

public sb_16bit_stereo_put_in_buffer_
sb_16bit_stereo_put_in_buffer_:
                test    ebx,ebx
                jz      sb16snone
                xor     eax,eax
                mov     ecx,dword ptr _auddmaptr
sb16sl:         mov     dx,word ptr [_left+2*eax]
                mov     word ptr [ecx+4*eax],dx
                mov     dx,word ptr [_right+2*eax]
                mov     word ptr 2[ecx+4*eax],dx
                inc     eax
                cmp     eax,ebx
                jne     sb16sl
                lea     ecx,[ecx+4*ebx]
                mov     dword ptr [_auddmaptr],ecx
sb16snone:      ret

; ebx - number of samples generated

public sb_16bit_mono_put_in_buffer_
sb_16bit_mono_put_in_buffer_:
                test    ebx,ebx
                jz      sb16mnone
                xor     eax,eax
                mov     ecx,dword ptr _auddmaptr
sb16ml:         mov     dx,word ptr [_left+2*eax]
                add     dx,word ptr [_right+2*eax]
                mov     word ptr [ecx+2*eax],dx
                inc     eax
                cmp     eax,ebx
                jne     sb16ml
                lea     ecx,[ecx+2*ebx]
                mov     dword ptr [_auddmaptr],ecx
sb16mnone:      ret

; ebx - number of samples generated

public sb_8bit_stereo_put_in_buffer_
sb_8bit_stereo_put_in_buffer_:
                test    ebx,ebx
                jz      sb8snone
                xor     eax,eax
                mov     ecx,dword ptr _auddmaptr
sb8sl:          mov     dx,word ptr [_left+2*eax]
                add     dh,80h
                mov     byte ptr [ecx+2*eax],dh
                mov     dx,word ptr [_right+2*eax]
                add     dh,80h
                mov     byte ptr 1[ecx+2*eax],dh
                inc     eax
                cmp     eax,ebx
                jne     sb8sl
                lea     ecx,[ecx+2*ebx]
                mov     dword ptr [_auddmaptr],ecx
sb8snone:       ret

; ebx - number of samples generated

public sb_8bit_mono_put_in_buffer_
sb_8bit_mono_put_in_buffer_:
                test    ebx,ebx
                jz      sb8mnone
                xor     eax,eax
                mov     ecx,dword ptr _auddmaptr
sb8ml:          mov     dx,word ptr [_left+2*eax]
                add     dx,word ptr [_right+2*eax]
                add     dh,80h
                mov     byte ptr [ecx+eax],dh
                inc     eax
                cmp     eax,ebx
                jne     sb8ml
                lea     ecx,[ecx+ebx]
                mov     dword ptr [_auddmaptr],ecx
sb8mnone:       ret



_TEXT ends

_DATA segment dword public 'DATA'
include data\sound.inc
include data\sblast.inc

_DATA ends

_BSS segment dword public 'DATA'
_BSS ends

end
